from django.db import models
import django

from . import forms
from .constants import DEFAULT_NUMBER_BITS
from .version import Version
from .utils import convert_version_int_to_string


class VersionField(models.Field):

    """
    A Field where version numbers are input/output as strings (e.g. 3.0.1)
    but stored in the db as converted integers for fast indexing
    """

    description = "A version number (e.g. 3.0.1)"

    def __init__(self, number_bits=DEFAULT_NUMBER_BITS, *args, **kwargs):
        self.number_bits = number_bits
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        """Use integer as internal representation."""
        return "integer"

    def to_python(self, value):
        if not value:
            return None

        if isinstance(value, Version):
            return value

        if isinstance(value, str):
            return Version(value, self.number_bits)

        return Version(
            convert_version_int_to_string(value, self.number_bits),
            self.number_bits
        )

    if django.VERSION < (2, 0):
        def from_db_value(self, value, expression, connection, context):
            return self.get_from_db_value(value, expression, connection)
    else:
        def from_db_value(self, value, expression, connection):
            return self.get_from_db_value(value, expression, connection)

    def get_from_db_value(self, value, expression, connection):
        """Convert data from database."""
        if value is None:
            return value
        return Version(
            convert_version_int_to_string(value, self.number_bits),
            self.number_bits)

    def get_prep_value(self, value):
        if isinstance(value, str):
            return int(Version(value, self.number_bits))

        if value is None:
            return None

        return int(value)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.VersionField,
            'number_bits': self.number_bits
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def __str__(self, value):
        return str(value)



import unittest

from django.test import TestCase
from django.db import models
from django.forms import ValidationError

from . import VersionField
from .constants import DEFAULT_NUMBER_BITS
from .version import Version
from . import forms


class DummyModel(models.Model):
    version = VersionField()


class VersionFieldTest(TestCase):

    def setUp(self):
        DummyModel.objects.create(version="0.1")
        DummyModel.objects.create(version="1.0")
        DummyModel.objects.create(version="1.0.1")

    def test_get_by_exact_version(self):
        thing = DummyModel.objects.get(version="0.1")
        self.assertEqual(thing.version, "0.1")
        self.assertEqual(thing.version, "0.1.0")

    def test_filter_by_greater_than_version(self):
        things = DummyModel.objects.filter(version__gt="0.1")
        self.assertEqual(len(things), 2)

        things = DummyModel.objects.filter(version__gt="1.0")
        self.assertEqual(len(things), 1)

        things = DummyModel.objects.filter(version__gt="1.0.1")
        self.assertEqual(len(things), 0)

    def test_filter_by_less_than_version(self):
        things = DummyModel.objects.filter(version__lt="0.1")
        self.assertEqual(len(things), 0)

        things = DummyModel.objects.filter(version__lt="1.0")
        self.assertEqual(len(things), 1)

        things = DummyModel.objects.filter(version__lt="1.0.1")
        self.assertEqual(len(things), 2)

    def test_overflow_number(self):
        error_occured = False
        try:
            DummyModel.objects.create(version="1.999.1")
        except ValueError:
            error_occured = True
        self.assertTrue(error_occured)

    def test_validate_positive(self):
        field = forms.VersionField()
        field.check_format("10.11.12")

    def test_validate_too_long(self):
        error_occured = False
        correct_error = False
        field = forms.VersionField()
        try:
            field.check_format("10.11.12.13")
        except ValidationError as e:
            error_occured = True
            if e.code == "too_long_version":
                correct_error = True
        self.assertTrue(error_occured)
        self.assertTrue(correct_error)

    def test_validate_not_numeric(self):
        error_occured = False
        correct_error = False
        field = forms.VersionField()
        try:
            field.check_format("10.x.1")
        except ValidationError as e:
            error_occured = True
            if e.code == "not_numeric_version":
                correct_error = True
        self.assertTrue(error_occured)
        self.assertTrue(correct_error)

    def test_validate_too_big(self):
        error_occured = False
        correct_error = False
        field = forms.VersionField()
        try:
            field.check_format("10.999.1")
        except ValidationError as e:
            error_occured = True
            if e.code == "version_component_too_big":
                correct_error = True
        self.assertTrue(error_occured)
        self.assertTrue(correct_error)


class DummyModelCustomBit(models.Model):
    version = VersionField(number_bits=(8, 16, 8))


class VersionFieldCustomBitsTest(TestCase):
    def setUp(self):
        DummyModelCustomBit.objects.create(version="1.999.1")

    def test_get_by_exact_version(self):
        thing = DummyModelCustomBit.objects.get(version="1.999.1")
        self.assertEqual(thing.version, "1.999.1")


class VersionObjectTestCase(unittest.TestCase):

    def test_equal_operator(self):
        self.assertEqual(
            Version("1.2.3", DEFAULT_NUMBER_BITS),
            Version("1.2.3", DEFAULT_NUMBER_BITS),
        )

    def test_lt_operator(self):
        self.assertTrue(
            Version("1.2.3", DEFAULT_NUMBER_BITS) <
            Version("1.2.4", DEFAULT_NUMBER_BITS)
        )
        self.assertFalse(
            Version("1.2.4", DEFAULT_NUMBER_BITS) <
            Version("1.2.3", DEFAULT_NUMBER_BITS)
        )

    def test_le_operator(self):
        self.assertTrue(
            Version("1.2.3", DEFAULT_NUMBER_BITS) <=
            Version("1.2.4", DEFAULT_NUMBER_BITS)
        )
        self.assertTrue(
            Version("1.2.4", DEFAULT_NUMBER_BITS) <=
            Version("1.2.4", DEFAULT_NUMBER_BITS)
        )

    def test_gt_operator(self):
        self.assertTrue(
            Version("1.2.4", DEFAULT_NUMBER_BITS) >
            Version("1.2.3", DEFAULT_NUMBER_BITS)
        )
        self.assertFalse(
            Version("2.3.4", DEFAULT_NUMBER_BITS) >
            Version("3.2.1", DEFAULT_NUMBER_BITS)
        )

    def test_ge_operator(self):
        self.assertTrue(
            Version("1.2.4", DEFAULT_NUMBER_BITS) >=
            Version("1.2.3", DEFAULT_NUMBER_BITS)
        )
        self.assertTrue(
            Version("3.3.3", DEFAULT_NUMBER_BITS) <=
            Version("3.3.3", DEFAULT_NUMBER_BITS)
        )



"""Custom form fields."""

from django import forms

from .version import Version
from .constants import DEFAULT_NUMBER_BITS
from .utils import convert_version_int_to_string


class VersionField(forms.CharField):

    """A form field dedicated to version numbers."""

    def __init__(self, number_bits=DEFAULT_NUMBER_BITS, **kwargs):
        self.number_bits = number_bits
        return super(VersionField, self).__init__(**kwargs)

    def check_format(self, string):
        """Check that value contains no more than N decimal numbers."""
        parts = string.split(".")
        actual_len = len(parts)
        allowed_len = len(self.number_bits)
        if actual_len > allowed_len:
            raise forms.ValidationError(
                "Version has %(actual)d components; only %(allowed)d "
                "components are allowed",
                code="too_long_version",
                params=dict(actual=actual_len, allowed=allowed_len))
        for i, (part, bits) in enumerate(zip(parts, self.number_bits), 1):
            if not part.isdigit():
                raise forms.ValidationError(
                    "Version's %(index)d component (%(part)s) is not numeric; "
                    "only numeric values are allowed",
                    code="not_numeric_version",
                    params=dict(index=i, part=part))
            num = int(part)
            max_allowed = (1 << bits) - 1
            if num > max_allowed:
                raise forms.ValidationError(
                    "Version's %(index)d component (%(part)s) is too big; "
                    "maximum allowed value for this component is %(allowed)d",
                    code="version_component_too_big",
                    params=dict(index=i, part=part, allowed=max_allowed))

    def to_python(self, value):
        """Verifies that value can be converted to a Version object."""
        if not value:
            return None

        self.check_format(value)

        if isinstance(value, str):
            return Version(value, self.number_bits)

        return Version(
            convert_version_int_to_string(value, self.number_bits),
            self.number_bits
        )



from django.conf import settings

DEFAULT_NUMBER_BITS = getattr(settings, 'VERSION_FIELD_NUMBER_BITS', (8, 8, 16))
DEFAULT_INPUT_SIZE  = getattr(settings, 'VERSION_FIELD_WIDGET_INPUT_SIZE', 10)
DEFAULT_VERSION_VALUE = getattr(settings, 'VERSION_FIELD_DEFAULT_VALUE', "0.0.0")


from django.forms.widgets import TextInput
from django.template.loader import render_to_string

from .constants import DEFAULT_INPUT_SIZE, DEFAULT_VERSION_VALUE


class VersionWidget(TextInput):

    class Media:
        css = {
            'all': ('django-versionfield/css/bootstrap.css',)
        }
        js = (
            'django-versionfield/js/jquery-3.5.1.min.js',
            'django-versionfield/js/bootstrap.min.js',
        )

    def render(self, name, value, attrs=None, **kwargs):

        # Get Attrs and set as read-only and reduce size
        attrs = attrs or {}
        if 'size' not in attrs:
            attrs['size'] = DEFAULT_INPUT_SIZE
        if value is None:
            value = DEFAULT_VERSION_VALUE

        # Get the default input html (as read-only text input field)
        input_html = super().render(name, value, attrs=attrs, **kwargs)

        # Render the Version Input widget
        rendered = render_to_string(
            'admin/django-versionfield/widgets/version-field.html',
            {
                'input_html': input_html,
                'value': value,
                'name': name,
                'is_required': self.is_required,
            },
        )
        return rendered



from .constants import DEFAULT_NUMBER_BITS
from .utils import convert_version_string_to_int, convert_version_int_to_string


class Version(object):
    def __init__(self, string, number_bits=DEFAULT_NUMBER_BITS):
        """
        Take in a verison string e.g. '3.0.1'
        Store it as a converted int
        """
        self.number_bits = number_bits
        self.internal_integer = convert_version_string_to_int(string, number_bits)

    def __str__(self):
        return str(convert_version_int_to_string(
                   self.internal_integer, self.number_bits))

    def __repr__(self):
        return self.__str__()

    def __int__(self):
        return self.internal_integer

    def __eq__(self, other):
        if not other:
            return False  # we are obviously a valid Version, but 'other' isn't
        if other == Ellipsis:
            return False # For pydantic use
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) == int(other)

    def __hash__(self):
        return int(self)

    def __lt__(self, other):
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) < int(other)

    def __le__(self, other):
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) <= int(other)

    def __gt__(self, other):
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) > int(other)

    def __ge__(self, other):
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) >= int(other)

    # Added Add + and Subtract - so 1.0.1 + 2.0.1 = 3.0.2 etc. Subtracting is max of 0.0.0
    def __add__(self, other):
        if not other:
            return self
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return Version(convert_version_int_to_string(int(self) + int(other), self.number_bits), self.number_bits)


    def __sub__(self, other):
        if not other:
            return self
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return Version(convert_version_int_to_string(max(int(self) - int(other), 0), self.number_bits), self.number_bits)
